@@ -5,20 +5,27 @@
 #include "config.h"
 /*
  * Prior to genericStack inclusion, we want to define a custom type for performance
+ * --------------------------------------------------------------------------------
  */
 #define GENERICSTACK_CUSTOM marpaESLIFValueResult_t
+#include <genericStack.h>
 
+/*
+ * Prior to genericHash inclusion, we define our hash size - a subject number
+ * --------------------------------------------------------------------------
+ */
 #ifndef MARPAESLIF_HASH_SIZE
 #define MARPAESLIF_HASH_SIZE 8 /* General default hash size */
 #endif
-
-#include <marpaWrapper.h>
-#include <genericStack.h>
 #include <genericHash.h>
-#include <genericLogger.h>
-#include <pcre2.h>
-#include <tconv.h>
 
+#include <marpaWrapper.h>                               /* Marpa engine */
+#include <genericLogger.h>                              /* Generic logger */
+#include <pcre2.h>                                      /* PCRE2 engine */
+#include <tconv.h>                                      /* Generic character/translator engine */
+
+/* Internal regex patterns */
+/* ----------------------- */
 #define INTERNAL_ANYCHAR_PATTERN "."                    /* This ASCII string is UTF-8 compatible */
 #define INTERNAL_UTF8BOM_PATTERN "\\x{FEFF}"            /* FEFF Unicode code point i.e. EFBBBF in UTF-8 encoding */
 #define INTERNAL_NEWLINE_PATTERN "(*BSR_UNICODE).*?\\R" /* newline as per unicode - we do .*? because our regexps are always anchored */
@@ -26,6 +33,8 @@
 #define INTERNAL_CHARACTERCLASSMODIFIERS_PATTERN "[eijmnsxDJUuaNubcA]+$"
 #define INTERNAL_REGEXMODIFIERS_PATTERN "[eijmnsxDJUuaNubcA]*$"
 
+/* Forward definitions */
+/* ------------------- */
 typedef struct  marpaESLIF_regex                 marpaESLIF_regex_t;
 typedef         marpaESLIFString_t               marpaESLIF_string_t;
 typedef enum    marpaESLIF_symbol_type           marpaESLIF_symbol_type_t;
@@ -49,9 +58,23 @@ typedef struct  marpaESLIF_lua_functioncall      marpaESLIF_lua_functioncall_t;
 typedef struct  marpaESLIF_lua_functiondecl      marpaESLIF_lua_functiondecl_t;
 typedef enum    marpaESLIF_json_type             marpaESLIF_json_type_t;
 typedef struct  marpaESLIF_pcre2_callout_context marpaESLIF_pcre2_callout_context_t;
+typedef struct  marpaESLIFGrammar_Lshare         marpaESLIFGrammar_Lshare_t;
 
 #include "marpaESLIF/internal/lua.h" /* For lua_State* */
 
+/* Implementations */
+/* --------------- */
+
+/*
+ * The general parent <-> child structure is:
+ *
+ * marpaESLIF
+ * -> marpaESLIFGrammar
+ *    -> marpaESLIF_grammar
+ *       -> marpaESLIFRecognizer that contains a marpaESLIF_stream that can be shared with other recognizers
+ *          -> marpaESLIFValue
+ */
+
 /* Internal event types that requires an action for faster lookup */
 typedef enum marpaESLIF_internal_event_action {
   MARPAESLIF_INTERNAL_EVENT_ACTION_NA = 0,
@@ -300,6 +323,7 @@ struct marpaESLIF_rule {
 /* A grammar */
 struct marpaESLIF_grammar {
   marpaESLIFGrammar_t   *marpaESLIFGrammarp;                 /* Shallow pointer to parent structure marpaESLIFGrammarp */
+  marpaESLIFGrammar_Lshare_t *Lsharep;                       /* Shallow pointer to parent structure's Lsharep - can never be NULL */
   int                    leveli;                             /* Grammar level */
   marpaESLIF_string_t   *descp;                              /* Grammar description */
   short                  descautob;                          /* True if alternative name was autogenerated */
@@ -346,53 +370,60 @@ enum marpaESLIF_json_type {
   _MARPAESLIF_JSON_TYPE_LAST
 };
 
+struct marpaESLIFGrammar_Lshare {
+  lua_State                 *L;                                 /* A Lua instance */
+  marpaESLIFRecognizer_t    *marpaESLIFRecognizerUnsharedTopp;  /* The unshared top-level recognizer that is running on this grammar */
+  marpaESLIFRecognizer_t    *marpaESLIFRecognizerLastInjectedp; /* The last marpaESLIFRecognizer injected */
+  marpaESLIFValue_t         *marpaESLIFValueLastInjectedp;      /* The last marpaESLIFValuep injected */
+};
+
 #define MARPAESLIFGRAMMARLUA_FOR_PARLIST 0
 #define MARPAESLIFGRAMMARLUA_FOR_EXPLIST 1
 struct marpaESLIF {
-  lua_State              *L;                           /* There is one lua state per ESLIF */
-  marpaESLIFGrammar_t    *marpaESLIFGrammarLuap;
-  marpaESLIFGrammar_t    *marpaESLIFGrammarLuapp[2];   /* C.f. MARPAESLIFGRAMMARLUA_FOR_PARLIST and MARPAESLIFGRAMMARLUA_FOR_EXPLIST */
-  marpaESLIFGrammar_t    *marpaESLIFGrammarp;          /* ESLIF has its own grammar -; */
-  marpaESLIFOption_t      marpaESLIFOption;
-  marpaESLIF_terminal_t  *anycharp;                    /* internal regex for match any character */
-  marpaESLIF_terminal_t  *newlinep;                    /* Internal regex for match newline */
-  marpaESLIFSymbol_t     *newlineSymbolp;              /* Internal symbol for match newline */
-  marpaESLIF_terminal_t  *stringModifiersp;            /* Internal regex for match string modifiers */
-  marpaESLIF_terminal_t  *characterClassModifiersp;    /* Internal regex for match character class modifiers */
-  marpaESLIF_terminal_t  *regexModifiersp;             /* Internal regex for match regex modifiers */
-  genericLogger_t        *traceLoggerp;                /* For cases where this is silent mode but compiled with TRACE */
-  short                   NULLisZeroBytesb;            /* An internal boolean to help when we can safely do calloc() */
-  short                   ZeroIntegerisZeroBytesb;     /* An internal boolean to help when we can safely do calloc() */
-  char                   *versions;                    /* Version */
-  int                     versionMajori;               /* Major version */
-  int                     versionMinori;               /* Minor version */
-  int                     versionPatchi;               /* Patch version */
-  marpaESLIFValueResult_t marpaESLIFValueResultTrue;   /* Pre-filled ::true value result */
-  marpaESLIFValueResult_t marpaESLIFValueResultFalse;  /* Pre-filled ::false value result */
+  marpaESLIFGrammar_t        *marpaESLIFGrammarLuap;
+  marpaESLIFGrammar_t        *marpaESLIFGrammarLuapp[2];   /* C.f. MARPAESLIFGRAMMARLUA_FOR_PARLIST and MARPAESLIFGRAMMARLUA_FOR_EXPLIST */
+  marpaESLIFGrammar_t        *marpaESLIFGrammarp;          /* ESLIF has its own grammar -; */
+  marpaESLIFOption_t          marpaESLIFOption;
+  marpaESLIF_terminal_t      *anycharp;                    /* internal regex for match any character */
+  marpaESLIF_terminal_t      *newlinep;                    /* Internal regex for match newline */
+  marpaESLIFSymbol_t         *newlineSymbolp;              /* Internal symbol for match newline */
+  marpaESLIF_terminal_t      *stringModifiersp;            /* Internal regex for match string modifiers */
+  marpaESLIF_terminal_t      *characterClassModifiersp;    /* Internal regex for match character class modifiers */
+  marpaESLIF_terminal_t      *regexModifiersp;             /* Internal regex for match regex modifiers */
+  genericLogger_t            *traceLoggerp;                /* For cases where this is silent mode but compiled with TRACE */
+  short                       NULLisZeroBytesb;            /* An internal boolean to help when we can safely do calloc() */
+  short                       ZeroIntegerisZeroBytesb;     /* An internal boolean to help when we can safely do calloc() */
+  char                       *versions;                    /* Version */
+  int                         versionMajori;               /* Major version */
+  int                         versionMinori;               /* Minor version */
+  int                         versionPatchi;               /* Patch version */
+  marpaESLIFValueResult_t     marpaESLIFValueResultTrue;   /* Pre-filled ::true value result */
+  marpaESLIFValueResult_t     marpaESLIFValueResultFalse;  /* Pre-filled ::false value result */
 #ifdef HAVE_LOCALE_H
-  struct lconv           *lconvp;
+  struct lconv               *lconvp;
 #endif
-  char                    decimalPointc;
-  const uint8_t          *tablesp;                     /* Output of pcre2_maketables */
+  char                        decimalPointc;
+  const uint8_t              *tablesp;                     /* Output of pcre2_maketables */
 #ifdef MARPAESLIF_HAVE_LONG_LONG
-  size_t                  llongmincharsl;              /* Number of digits of LLONG_MIN */
-  size_t                  llongmaxcharsl;              /* Number of digits of LLONG_MAX */
+  size_t                      llongmincharsl;              /* Number of digits of LLONG_MIN */
+  size_t                      llongmaxcharsl;              /* Number of digits of LLONG_MAX */
 #else
-  size_t                  longmincharsl;               /* Number of digits of LONG_MIN */
-  size_t                  longmaxcharsl;               /* Number of digits of LONG_MAX */
+  size_t                      longmincharsl;               /* Number of digits of LONG_MIN */
+  size_t                      longmaxcharsl;               /* Number of digits of LONG_MAX */
 #endif
 #ifdef MARPAESLIF_INFINITY
-  float                   positiveinfinityf;           /* +Inf */
-  float                   negativeinfinityf;           /* -Inf */
+  float                       positiveinfinityf;           /* +Inf */
+  float                       negativeinfinityf;           /* -Inf */
 #endif
 #ifdef MARPAESLIF_NAN
-  float                   positivenanf;                /* +NaN */
-  float                   negativenanf;                /* -NaN */
-  short                   nanconfidenceb;              /* 1 when ESLIF think that NaN representation is correct */
+  float                       positivenanf;                /* +NaN */
+  float                       negativenanf;                /* -NaN */
+  short                       nanconfidenceb;              /* 1 when ESLIF think that NaN representation is correct */
 #endif
   /* For JSON grammars : the symbols that depend on strictness */
-  marpaESLIF_symbol_t    *jsonStringpp[_MARPAESLIF_JSON_TYPE_LAST];
-  marpaESLIF_symbol_t    *jsonConstantOrNumberpp[_MARPAESLIF_JSON_TYPE_LAST];
+  marpaESLIF_symbol_t        *jsonStringpp[_MARPAESLIF_JSON_TYPE_LAST];
+  marpaESLIF_symbol_t        *jsonConstantOrNumberpp[_MARPAESLIF_JSON_TYPE_LAST];
+  marpaESLIFGrammar_Lshare_t  Lshare;                  /* A Lua instance, used by all sub-grammars of ESLIF */
 };
 
 struct marpaESLIFGrammar {
@@ -420,8 +451,14 @@ struct marpaESLIFGrammar {
   /* For JSON grammars : the symbols that depend on strictness */
   marpaESLIF_symbol_t       *jsonStringp; /* Shallow pointer */
   marpaESLIF_symbol_t       *jsonConstantOrNumberp; /* Shallow pointer */
-  lua_State                 *L;                  /* A Lua thread */
-  int                        L_r;                /* Reference to the lua thread in the main thread */
+
+  /* Lua singleton. There is a difficulty with generated grammars: we do not want to attach them  */
+  /* to the original grammar, because they are volatile. This is why the L singleton and          */
+  /* associated singleton constants are all in a specific structure, and they are accessible      */
+  /* only via a pointer.                                                                          */
+  /* By definition a grammar is the owner of the singleton if *Lsharep == &_Lshare.               */
+  marpaESLIFGrammar_Lshare_t _Lshare;
+  marpaESLIFGrammar_Lshare_t *Lsharep;
 };
 
 struct marpaESLIF_meta {
@@ -452,6 +489,7 @@ struct marpaESLIF_stringGenerator {
 
 struct marpaESLIFValue {
   marpaESLIF_t                *marpaESLIFp;
+  marpaESLIFGrammar_Lshare_t  *Lsharep;                       /* Shallow pointer to parent structure's Lsharep - can never be NULL */
   marpaESLIFRecognizer_t      *marpaESLIFRecognizerp;
   marpaESLIFValueOption_t      marpaESLIFValueOption;
   marpaWrapperValue_t         *marpaWrapperValuep;
@@ -462,7 +500,6 @@ struct marpaESLIFValue {
   short                        inValuationb;
   marpaESLIF_symbol_t         *symbolp;
   marpaESLIF_rule_t           *rulep;
-  marpaESLIFValue_t           *marpaESLIFValueLastInjectedp;
   char                        *actions; /* Shallow pointer to action "name", depends on action type */
   marpaESLIF_action_t         *actionp; /* Shallow pointer to action */
   marpaESLIF_string_t         *stringp; /* Not NULL only when is a literal - then callback is forced to be internal */
@@ -505,12 +542,9 @@ struct marpaESLIFRecognizer {
   /* except in very precise situations (typically the new()/free() or when */
   /* faking a new() method). */
   marpaESLIF_t                *marpaESLIFp;
-
-  /* Because recognizers can be cached we cannot afford marpaESLIFGrammarp to point to something on the stack when we free: Grammarp and grammarp */
-  /* We guarantee that pointers below grammarp are correct (these are shallow pointers) */
-  marpaESLIFGrammar_t          _marpaESLIFGrammar;
-  marpaESLIF_grammar_t         _grammar;
-  marpaESLIFGrammar_t         *marpaESLIFGrammarp;
+  marpaESLIF_grammar_t        *grammarp;
+  marpaESLIFGrammar_Lshare_t *Lsharep;                       /* Shallow pointer to parent structure's Lsharep - TAKE CARE - can be NULL because sometimes grammarp can be NULL */
+  marpaESLIFGrammar_t        *marpaESLIFGrammarp;            /* Shallow pointer to parent structure's marpaESLIFGrammarp - TAKE CARE - can be NULL because sometimes grammarp can be NULL */
   
   marpaESLIFRecognizerOption_t marpaESLIFRecognizerOption;
   marpaWrapperRecognizer_t    *marpaWrapperRecognizerp; /* Current recognizer */
@@ -566,11 +600,6 @@ struct marpaESLIFRecognizer {
   int                          numberOfStartCompletionsi; /* Computed only if maxStartCompletionsi != 0 */
   size_t                       lastSizeBeforeCompletionl; /* Computed only if maxStartCompletionsi is != 0 */
 
-  /* When a recognizer needs to discard, we prepare storage for marpaESLIFGrammarp, marpaESLIFRecognizerOptionp and marpaESLIFValueOptionp. */
-  /* This a lazy initialization, driven by the grammarDiscardInitializedb flag. */
-  short                        grammarDiscardInitializedb;
-  marpaESLIFGrammar_t          marpaESLIFGrammarDiscard;
-  marpaESLIF_grammar_t         grammarDiscard;
   marpaESLIFRecognizerOption_t marpaESLIFRecognizerOptionDiscard;
   marpaESLIFValueOption_t      marpaESLIFValueOptionDiscard;
 
@@ -588,9 +617,6 @@ struct marpaESLIFRecognizer {
   char                        *lastDiscards;    /* Bytes */
 
   /* For lua action callbacks */
-  marpaESLIFRecognizer_t      *marpaESLIFRecognizerLastInjectedp;
-  lua_State                   *L;              /* A Lua thread - only the top-level recognizer owns it */
-  int                          L_r;            /* Reference to the lua thread in the main thread */
   char                        *actions;        /* Shallow pointer to action "name", depends on action type */
   marpaESLIF_action_t         *actionp;        /* Shallow pointer to action */
 
@@ -626,6 +652,8 @@ struct marpaESLIFRecognizer {
   marpaESLIFAction_t             *getContextActionp;  /* Getting the context is a common function that is stored at recognizer level */
   marpaESLIFAction_t             *setContextActionp;  /* Setting the context is a common function that is stored at recognizer level */
   marpaESLIFAction_t             *popContextActionp;  /* Getting the context is a common function that is stored at recognizer level */
+
+  marpaESLIFRecognizer_t         *marpaESLIFRecognizerSharedp;
 };
 
 struct marpaESLIF_symbol_data {
